<?php

/**
 * @file
 * Builds placeholder replacement tokens for iterating.
 */

use Drupal\Core\Render\BubbleableMetadata;
use Drupal\mustache\Render\IterableMarkup;

/**
 * Implements hook_token_info().
 */
function mustache_token_token_info() {
  $info = [];
  // Iterateable tokens.
  $info['types']['iterate'] = [
    'name' => t('Iterate'),
    'description' => t('Tokens for iterable data. You may use <a href="https://mustache.github.io/mustache.5.html" target="_blank" rel="noopener noreferrer">Mustache syntax</a>. Example: <em>&lt;h1&gt;To Do&lt;&#47;h1&gt;&lt;ul&gt;{{#iterate.loop.node.field_mytodolist}}&lt;li&gt;{{#first}}First step: {{/first}}{{^first}}Next step no. {{position}}: {{/first}}{{.}}{{#last}}&lt;&#47;li&gt;&lt;li&gt;... and take care of yourself.{{/last}}&lt;&#47;li&gt;{{/iterate.loop.node.field_mytodolist}}&lt;&#47;ul&gt;</em>'),
  ];
  $info['tokens']['iterate']['loop'] = [
    'name' => t('Loop'),
    'description' => t('Loop through iterable data.'),
    'dynamic' => TRUE,
  ];
  $info['tokens']['iterate']['first'] = [
    'name' => t('First item'),
    'description' => t('The first item of the iterable data.'),
    'dynamic' => TRUE,
  ];
  $info['tokens']['iterate']['last'] = [
    'name' => t('Last item'),
    'description' => t('The last item of the iterable data.'),
    'dynamic' => TRUE,
  ];
  $info['tokens']['iterate']['count'] = [
    'name' => t('Count'),
    'description' => t('Count the items of iterable data.'),
    'dynamic' => TRUE,
  ];
  $info['tokens']['iterate']['empty'] = [
    'name' => t('Empty'),
    'description' => t('Whether the iterable data is empty ("Empty") or not ("Not empty").'),
    'dynamic' => TRUE,
  ];
  $info['tokens']['iterate']['sum'] = [
    'name' => t('Sum'),
    'description' => t('Sum of iterable data values.'),
    'dynamic' => TRUE,
  ];
  $info['tokens']['iterate']['min'] = [
    'name' => t('Minimum'),
    'description' => t('The minimum of iterable data values.'),
    'dynamic' => TRUE,
  ];
  $info['tokens']['iterate']['max'] = [
    'name' => t('Maximum'),
    'description' => t('The maximum of iterable data values.'),
    'dynamic' => TRUE,
  ];
  $info['tokens']['iterate']['avg'] = [
    'name' => t('Average'),
    'description' => t('The average of iterable data values.'),
    'dynamic' => TRUE,
  ];
  return $info;
}

/**
 * Implements hook_tokens().
 */
function mustache_token_tokens($type, $tokens, array $data, array $options, BubbleableMetadata $bubbleable_metadata) {
  $replacements = [];

  if ($type == 'iterate') {
    /** @var \Drupal\mustache\MustacheTokenIterate $token_iterate */
    $token_iterate = \Drupal::service('mustache.token_iterate');
    foreach ($tokens as $name => $original) {
      $token_keys = explode(':', $name);
      $op = array_shift($token_keys);
      if (empty($token_keys)) {
        continue;
      }
      $target = $token_iterate->getIterableTarget($token_keys, $data, $options, $bubbleable_metadata);
      switch ($op) {

        case 'loop':
          $replacements[$original] = IterableMarkup::create($target);
          break;

        case 'first':
          if ($first = reset($target)) {
            $replacements[$original] = $first;
          }
          break;

        case 'last':
          if ($last = end($target)) {
            $replacements[$original] = $last;
          }
          break;

        case '_count':
        case 'count':
          $replacements[$original] = count($target);
          break;

        case 'empty':
          $replacements[$original] = empty($target) ? t('Empty') : t('Not empty');
          break;

        case 'avg':
        case 'sum':
          $sum = 0.0;
          foreach ($target as $value) {
            if (is_numeric($value)) {
              $sum += (float) $value;
            }
            elseif (is_string($value)) {
              $sum += strlen($value);
            }
            elseif (is_array($value) || $value instanceof \Countable) {
              $sum += count($value);
            }
            elseif (!empty($value)) {
              $sum++;
            }
          }
          if (floor($sum) == $sum) {
            $sum = (int) $sum;
          }
          if ($op == 'avg' && ($count = count($target))) {
            $sum /= $count;
            $sum = number_format($sum, 2);
          }
          $replacements[$original] = $sum;
          break;

        case 'min':
          $min = NULL;
          foreach ($target as $value) {
            if (is_numeric($value)) {
              $min = $min === NULL || $min > $value ? $value : $min;
            }
            elseif (is_string($value)) {
              $min = $min === NULL || $min > strlen($value) ? strlen($value) : $min;
            }
            elseif (is_array($value) || $value instanceof \Countable) {
              $min = $min === NULL || $min > count($value) ? count($value) : $min;
            }
          }
          if (floor($min) == $min) {
            $min = (int) $min;
          }
          if ($min !== NULL) {
            $replacements[$original] = $min;
          }
          break;

        case 'max':
          $max = NULL;
          foreach ($target as $value) {
            if (is_numeric($value)) {
              $max = $max === NULL || $max < $value ? $value : $max;
            }
            elseif (is_string($value)) {
              $max = $max === NULL || $max < strlen($value) ? strlen($value) : $max;
            }
            elseif (is_array($value) || $value instanceof \Countable) {
              $max = $max === NULL || $max < count($value) ? count($value) : $max;
            }
          }
          if (floor($max) == $max) {
            $max = (int) $max;
          }
          if ($max !== NULL) {
            $replacements[$original] = $max;
          }
          break;

      }
    }
  }

  return $replacements;
}
