<?php

namespace Drupal\mustache\Helpers;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Core\Access\AccessibleInterface;
use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\EntityReferenceFieldItemListInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\TypedData\ListInterface;
use Drupal\Core\TypedData\TypedDataInterface;
use Drupal\mustache\Render\IterableMarkup;
use Drupal\mustache\Render\Markup;

/**
 * A helper object to introspect available data within Mustache templates.
 *
 * Usage:
 *
 * You can use the "magical" {{debug}} variable anywhere Mustache template
 * syntax is allowed. For example, the Mustache input filter plugin or - when
 * "mustache_token" is installed - any input that supports tokens.
 * When that debug variable is used in a template, the contents will not be
 * printed out in the rendered template, but instead printed out as a message
 * on the currently opened page.
 *
 * Printed debug messages are only shown for users who have the
 * "view mustache debug messages" permission.
 *
 * @code
 * {{debug.keys}}
 * @endcode
 */
class MustacheDebug implements MarkupInterface {

  /**
   * The attached data.
   *
   * @var mixed
   */
  protected $data;

  /**
   * The element as render array.
   *
   * @var array
   */
  protected $element;

  /**
   * The messenger, if any.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface|null
   */
  protected $messenger;

  /**
   * The MustacheDebug constructor.
   *
   * @param mixed &$data
   *   The attached data.
   * @param array &$element
   *   The element as render array.
   * @param \Drupal\Core\Messenger\MessengerInterface|null $messenger
   *   The messenger, if any.
   */
  public function __construct(&$data, array &$element, MessengerInterface $messenger = NULL) {
    $this->data = &$data;
    $this->element = &$element;
    $this->messenger = $messenger;
  }

  /**
   * {@inheritdoc}
   */
  public function __toString() {
    \Drupal::logger('mustache')->warning(sprintf("A {{debug}} variable is set in a Mustache template (%s). Remove that variable from the template when it is not needed anymore.", $this->element['#template']['name']));
    return "";
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return NULL;
  }

  /**
   * Prints out a list of available data keys.
   *
   * @return string
   *   An empty string is returned, so that rendered template output is not
   *   affected when using a {{debug}} variable.
   */
  public function keys() {
    \Drupal::logger('mustache')->warning(sprintf("A {{debug.keys}} variable is set in a Mustache template (%s). Remove that variable from the template so that these messages will not show up anymore.", $this->element['#template']['name']));
    if (!$this->messenger) {
      return "";
    }

    $module_handler = \Drupal::moduleHandler();
    $with_tokens = isset($this->element['#with_tokens']) && $this->element['#with_tokens'] !== FALSE;

    $keys = [];
    if (is_iterable($this->data)) {
      foreach ($this->data as $key => $value) {
        if (empty($key) || $key === 'debug') {
          continue;
        }
        $this->buildKeysInfo($keys, $value, $key);
      }
    }
    if ($with_tokens) {
      $token_settings = &$this->element['#with_tokens'];
      $token_data = isset($token_settings['data']) ? $token_settings['data'] : [];
      /** @var \Drupal\mustache\MustacheTokenProcessor $token_processor */
      $token_processor = \Drupal::service('mustache.token_processor');
      $token_processor->processData($token_data, 'view');
      if (!empty($token_data)) {
        $token_info = \Drupal::token()->getInfo();
        foreach ($token_info['tokens'] as $type => $tokens) {
          if (isset($token_data[$type])) {
            $this->buildKeysInfo($keys, $token_data[$type], $type);
            foreach ($tokens as $key => $token) {
              $key = $type . '.' . str_replace(':', '.', $key);
              $keys[$key] = '<b>{{' . $key . '}}</b> <em>' . $token['name'] . '</em>';
            }
          }
        }
      }
    }
    foreach (array_keys($keys) as $key) {
      $key_parts = explode('.', $key);
      if (is_numeric(end($key_parts))) {
        // Encourage the usage concrete property names. This also lowers the
        // possibly huge amount of suggestions, at least for a little bit.
        unset($keys[$key]);
      }
    }
    if (!empty($keys)) {
      $list = Markup::create('<ul><li>' . implode('</li><li>', $keys) . '</li></ul>');
      $this->messenger->addMessage(t('Available data keys:<br />@list', ['@list' => $list]));
      if ($module_handler->moduleExists('mustache_token')) {
        $this->messenger->addMessage(t('If you want to iterate through any of the above mentioned keys, you may prefix the key with "iterate.&lt;operation&gt;.*", for example {{iterate.loop.node.field_mytodolist}}. Available <b>&lt;operation&gt;</b> keys:<br /><ul><li>{{iterate.<b>loop</b>.*}}</li><li>{{iterate.<b>first</b>.*}}</li><li>{{iterate.<b>last</b>.*}}</li><li>{{iterate.<b>count</b>.*}}</li><li>{{iterate.<b>empty</b>.*}}</li><li>{{iterate.<b>sum</b>.*}}</li><li>{{iterate.<b>min</b>.*}}</li><li>{{iterate.<b>max</b>.*}}</li><li>{{iterate.<b>avg</b>.*}}</li></ul>'));
      }
    }
    elseif (!$with_tokens) {
      $this->messenger->addWarning(t('No explicitly set data keys are available.'));
    }
    if ($with_tokens) {
      if ($module_handler->moduleExists('token') && $module_handler->moduleExists('help')) {
        $this->messenger->addMessage(t('<a href="/admin/help/token" target="_blank">Global tokens</a> may be available for you. Available tokens like "[site:name]" may be used as Mustache variables like "{{site.name}}".'));
      }
      else {
        $this->messenger->addMessage(t('Global tokens may be available for you. Available tokens like "[site:name]" may be used as Mustache variables like "{{site.name}}".'));
      }
    }
    $this->messenger->addWarning(t('Above messages are printed because a {{debug.keys}} variable has been set in a Mustache template. Remove that variable from the template so that these messages will not show up anymore.'));

    return "";
  }

  /**
   * Implements the magic __serialize() method.
   *
   * @return array
   *   The array for serialization.
   */
  public function __serialize(): array {
    return [];
  }

  /**
   * Recursively builds up the keys info for a given value.
   *
   * @param array &$keys
   *   Already added keys.
   * @param mixed $data
   *   The currently scoped data value.
   * @param mixed $parent_key
   *   The parent key.
   * @param int $level
   *   (Optional) The current nesting level. This method aborts at level 3.
   * @param array $additional_info
   *   (Optional) Additional info.
   */
  protected function buildKeysInfo(array &$keys, $data, $parent_key, $level = 0, array $additional_info = []) {
    if ($level > 3) {
      return;
    }
    $level++;

    $is_iterable = is_iterable($data);
    if ($data instanceof EntityReferenceFieldItemListInterface) {
      $target_type = $data->getFieldDefinition()->getSetting('target_type');
      $entities = $data->referencedEntities();
      $count = count($entities);
      $cardinality = (int) $data->getFieldDefinition()->getFieldStorageDefinition()->getCardinality();
      foreach ($entities as $delta => $entity) {
        $additional_info = [];
        if ($count > 1) {
          $additional_info = [t('available items: <b>@num</b>', ['@num' => $count])];
        }
        $this->buildKeysInfo($keys, $entity, $parent_key . '.' . $target_type, $level, $additional_info);
        if ($cardinality !== 1) {
          $additional_info = [];
          if ($count > 1) {
            $additional_info = [t('available &lt;delta&gt; keys: <b>@keys</b>', ['@keys' => implode(',', array_keys($entities))])];
            $this->buildKeysInfo($keys, $entity, $parent_key . '.&lt;delta&gt;' . '.' . $target_type, $level, $additional_info);
          }
          else {
            $this->buildKeysInfo($keys, $entity, $parent_key . '.' . $delta . '.' . $target_type, $level, $additional_info);
          }
        }
      }
    }
    elseif ($data instanceof EntityReferenceItem) {
      $target_type = $data->getFieldDefinition()->getSetting('target_type');
      $item_value = $data->getValue();
      if (isset($item_value['target_id']) || isset($item_value['entity'])) {
        $entity = !empty($item_value['entity']) ? $item_value['entity'] : \Drupal::entityTypeManager()->getStorage($target_type)->load($item_value['target_id']);
        if ($entity) {
          $this->buildKeysInfo($keys, $entity, $parent_key . '.' . $target_type, $level);
        }
      }
    }
    if ($data instanceof ListInterface) {
      $count = $data->count();
      $cardinality = $data instanceof FieldItemListInterface ? $data->getFieldDefinition()->getFieldStorageDefinition()->getCardinality() : $count;
      foreach ($data as $delta => $item) {
        $additional_info = [];
        if ($count > 1) {
          $additional_info = [t('available items: <b>@num</b>', ['@num' => $count])];
        }
        $this->buildKeysInfo($keys, $item, $parent_key, $level, $additional_info);
        if ($cardinality !== 1) {
          $additional_info = [];
          if ($count > 1) {
            $additional_info = [t('available <b>&lt;delta&gt;</b> keys: <b>@keys</b>', ['@keys' => implode(',', array_keys($data->getValue()))])];
            $this->buildKeysInfo($keys, $item, $parent_key . '.&lt;delta&gt;', $level, $additional_info);
          }
          else {
            $this->buildKeysInfo($keys, $item, $parent_key . '.' . $delta, $level, $additional_info);
          }
        }
        return;
      }
    }
    if (!$is_iterable) {
      $data = [$parent_key => $data];
    }
    foreach ($data as $key => $value) {
      $key = $is_iterable ? implode('.', [$parent_key, $key]) : $key;
      if ($value instanceof IterableMarkup) {
        // Exclude "background noise" that comes from our beloved
        // MustacheTokenProcessor, which generates a stringable data tree.
        continue;
      }
      if (($value instanceof AccessibleInterface) && !$value->access('view')) {
        continue;
      }
      if ($value instanceof FieldableEntityInterface) {
        // From here on, work with the entity wrapped as typed data object.
        $value = $value->getTypedData();
      }
      if ($value instanceof TypedDataInterface) {
        $data_definition = $value->getDataDefinition();
        if (!($data_desc = (string) $data_definition->getLabel())) {
          $data_type = explode(':', $data_definition->getDataType());
          $data_desc = end($data_type);
        }
        $data_desc = !empty($data_desc) ? [$data_desc] : [];

        $keys[$key] = '<b>{{' . $key . '}}</b> <em>' . implode(', ', $data_desc) . '</em>';
      }
      elseif ($value instanceof ConfigEntityInterface) {
        // Some config objects don't give any insight about their properties,
        // thus work with the array representation from here instead.
        $keys[$key] = '<b>{{' . $key . '}}</b> <em>' . $value->getEntityType()->getLabel() . ' (' . t('not a scalar value') . ')</em>';
        $value = $value->toArray();
      }
      else {
        $keys[$key] = '<b>{{' . $key . '}}</b> <em>' . gettype($value) . '</em>';
      }
      if (!empty($additional_info)) {
        $keys[$key] .= ', ' . implode(', ', $additional_info);
      }
      if (is_iterable($value) || $value instanceof TypedDataInterface) {
        $this->buildKeysInfo($keys, $value, $key, $level, $additional_info);
      }
    }
  }

}
