<?php

namespace Drupal\mustache;

use Drupal\Core\Access\AccessibleInterface;
use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\Context\ContextRepositoryInterface;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\Utility\Token;
use Drupal\mustache\Render\IterableMarkup;

/**
 * Takes care of token processing within the render pipeline.
 *
 * @see \Drupal\mustache\Element\Mustache
 */
class MustacheTokenProcessor {

  /**
   * The cache bin to store tokenized template content.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * The token service.
   *
   * @var Drupal\Core\Utility\Token
   */
  protected $token;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The context repository.
   *
   * @var \Drupal\Core\Plugin\Context\ContextRepositoryInterface
   */
  protected $contextRepository;

  /**
   * Statically cached context IDs.
   *
   * @var string[]
   */
  protected static $contextIds;

  /**
   * The MustacheTokenProcessor constructor.
   *
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache bin to store tokenized template content.
   * @param \Drupal\Core\Utility\Token $token
   *   The token service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Plugin\Context\ContextRepositoryInterface $context_repository
   *   The context repository.
   */
  public function __construct(CacheBackendInterface $cache, Token $token, EntityTypeManagerInterface $entity_type_manager, ContextRepositoryInterface $context_repository) {
    $this->cache = $cache;
    $this->token = $token;
    $this->entityTypeManager = $entity_type_manager;
    $this->contextRepository = $context_repository;
  }

  /**
   * Processes tokens for the given template content.
   *
   * Before invoking this method, check before whether it is really needed.
   *
   * @param array &$element
   *   The current element build array.
   * @param string &$template_content
   *   The current template content.
   *
   * @\Drupal\mustache\Element\Mustache::processTokens().
   */
  public function processTokens(array &$element, &$template_content) {
    if (!isset($element['#data'])) {
      $element['#data'] = [];
    }
    if (!is_array($element['#with_tokens'])) {
      $element['#with_tokens'] = [];
    }

    $token_service = $this->token;
    $token_settings = &$element['#with_tokens'];
    $token_options = isset($token_settings['options']) ? $token_settings['options'] : [];

    // By default, empty or non-matched variables shall be cleared.
    if (!isset($token_options['clear'])) {
      $token_options['clear'] = TRUE;
    }

    if (!empty($element['#cache'])) {
      $bubbleable_metadata = BubbleableMetadata::createFromRenderArray($element);
    }
    else {
      $bubbleable_metadata = new BubbleableMetadata();
    }

    // Build up the token data.
    $token_data = isset($token_settings['data']) ? $token_settings['data'] : [];
    $this->processData($token_data, 'view', $bubbleable_metadata);

    $cid = 'mustache:tokenized:' . hash('crc32b', $template_content);
    if ($cached = $this->cache->get($cid)) {
      $template_content = $cached->data['content'];
      $template_tokens = $cached->data['tokens'];
      $mustache_variables = $cached->data['variables'];
    }
    else {
      if (strpos($template_content, 'count') !== FALSE) {
        // Special handling for count, because IterableMarkup implements
        // the count method for the \Countable interface, but at the same
        // time it may have an item that is called "count" too.
        $template_content = str_replace('iterate:count:', 'iterate:_count:', $template_content);
        $template_content = str_replace('iterate.count.', 'iterate._count.', $template_content);
      }

      // Scan for regular tokens.
      $template_tokens = $token_service->scan($template_content);
      // Also include Mustache variables as tokens.
      preg_match_all('/\{\{[^a-zA-Z\d\_\-\{\#\&\^]*([a-zA-Z\d\_\-\?\<\>\.]+)[^a-zA-Z\d\_\-\.\}]*\}\}/x', $template_content, $matches);
      if (!empty($matches[1])) {
        foreach ($matches[1] as $match) {
          $match = trim($match);
          $parts = explode('.', $match);
          $type = array_shift($parts);
          if (empty($parts)) {
            // A token always consists of at least two parts: The token type as
            // first prefix, followed by a ":" and a name (e.g. node:title).
            // Therefore, we do not try to include variables that only consist
            // of one key (i.e. only include variables with a "." dot).
            continue;
          }
          if (!isset($token_data[$type])) {
            $type_hyphened = str_replace('_', '-', $type);
            if (isset($token_data[$type_hyphened])) {
              $template_content = str_replace($match, $type_hyphened . '.' . implode('.', $parts), $template_content);
              $type = $type_hyphened;
            }
          }
          $name = implode(':', $parts);
          array_unshift($parts, $type);
          $token = '[' . implode(':', $parts) . ']';
          $template_tokens[$type][$name] = $token;
        }
      }
      if (empty($template_tokens)) {
        return;
      }
      unset($matches);

      // Transform tokens into Mustache variables.
      $mustache_variables = [];
      foreach ($template_tokens as $type => $tokens) {
        foreach ($tokens as $token) {
          $variable = str_replace(':', '.', substr($token, 1, -1));
          if (strpos($variable, '-') !== FALSE) {
            // Transform any hyphens into underscores, so that built up
            // data trees always match for given keys, even when hyphens and
            // underscores are being mixed up.
            $variable_underscore = str_replace('-', '_', $variable);
            $template_content = str_replace($variable, $variable_underscore, $template_content);
            $variable = $variable_underscore;
          }
          $mustache_variables[$token] = $variable;
        }
      }

      // Ensure that standalone tokens are being wrapped by Mustache brackets,
      // and replace all token versions by their Mustache variable counterparts.
      preg_match_all('/(\{\{)?[^a-zA-Z\d\_\-\{\#\&\^]*(\[[^\s\[\]:]+:[^\[\]]+\])[^a-zA-Z\d\_\-\?\.\}]*(\}\})?/x', $template_content, $matches);
      if (!empty($matches[2])) {
        foreach ($matches[2] as $i => $token) {
          if (!isset($mustache_variables[$token])) {
            continue;
          }
          $search = $matches[0][$i];
          $variable = $mustache_variables[$token];
          if ($matches[1][$i] !== '{{') {
            $variable = '{{' . $variable;
          }
          if ($matches[3][$i] !== '}}') {
            $variable .= '}}';
          }
          $replace = str_replace($token, $variable, $search);
          $template_content = str_replace($search, $replace, $template_content);
        }
      }

      // Cache the tokenized results for the next request.
      $this->cache->set($cid, [
        'content' => $template_content,
        'tokens' => $template_tokens,
        'variables' => $mustache_variables,
      ]);
    }

    // Let the token system generate replacement values.
    $replacements = [];
    foreach ($template_tokens as $type => $tokens) {
      $replacements += $token_service->generate($type, $tokens, $token_data, $token_options, $bubbleable_metadata);
      if ($type !== 'debug' && $type !== 'iterate') {
        // In case no replacement value could be generated, try to receive
        // a fallback value by using the token iterate service.
        foreach ($tokens as $token) {
          if (!isset($replacements[$token]) && ($target = static::getTokenIterate()->getIterableTarget(substr($token, 1, -1), $token_data, $token_options, $bubbleable_metadata))) {
            $replacements[$token] = IterableMarkup::create($target);
          }
        }
      }
      if ($token_options['clear']) {
        $replacements += array_fill_keys($tokens, '');
      }
    }

    // No explicit escaping of token values is being done here, because
    // wrapped Mustache variables will be escaped by default, if not using
    // the triple {{{}}} parenthesis.
    // Optionally alter the list of replacement values.
    if (!empty($token_options['callback'])) {
      $function = $token_options['callback'];
      $function($replacements, $token_data, $token_options, $bubbleable_metadata);
    }

    $bubbleable_metadata->applyTo($element);

    // Make the replacement values accessible as an iterable data tree.
    $mustache_data = IterableMarkup::create();
    foreach ($replacements as $token => $replacement) {
      $variable = $mustache_variables[$token];

      $iterable = &$mustache_data;
      foreach (explode('.', $variable) as $parent) {
        if (!isset($iterable[$parent])) {
          $iterable[$parent] = IterableMarkup::create([], NULL, $iterable);
        }
        $iterable = &$iterable[$parent];
      }

      if (is_iterable($replacement)) {
        foreach ($replacement as $key => $value) {
          if (is_numeric($key)) {
            $iterable[] = IterableMarkup::create([], $value, $iterable);
          }
          else {
            $iterable[$key] = IterableMarkup::create([], $value, $iterable);
          }
        }
      }
      elseif (is_scalar($replacement) || (is_object($replacement) && method_exists($replacement, '__toString'))) {
        $iterable->value = $replacement;
      }
      // Unset to not manipulate the last reference in the next loop.
      unset($iterable);
    }

    if (empty($element['#data'])) {
      $element['#data'] = $mustache_data;
    }
    elseif (is_array($element['#data'])) {
      foreach ($element['#data'] as $key => $value) {
        if (is_scalar($value) || (is_object($value) && method_exists($value, '__toString'))) {
          $mustache_data[$key] = $value;
        }
      }
      $element['#data'] = $mustache_data;
    }
    elseif (isset($element['#override_data']) && is_array($element['#override_data'])) {
      foreach ($element['#override_data'] as $key => $value) {
        if (is_scalar($value) || (is_object($value) && method_exists($value, '__toString'))) {
          $mustache_data[$key] = $value;
        }
      }
      $element['#override_data'] = $mustache_data;
    }
    else {
      $element['#override_data'] = $mustache_data;
    }
  }

  /**
   * Processes given token data by enriching with context and access filtering.
   *
   * @param array &$token_data
   *   The token data to process.
   * @param string $check_access
   *   (Optional) The access check to perform. Set NULL to skip access filter.
   * @param \Drupal\Core\Render\BubbleableMetadata|null $bubbleable_metadata
   *   (Optional) The object to collect bubbleable metadata.
   */
  public function processData(array &$token_data, $check_access = 'view', $bubbleable_metadata = NULL) {
    if (!isset($bubbleable_metadata)) {
      $bubbleable_metadata = new BubbleableMetadata();
    }

    /** @var \Drupal\token\TokenEntityMapperInterface $entity_token_type_mapper */
    $entity_token_type_mapper = \Drupal::hasService('token.entity_mapper') ? \Drupal::service('token.entity_mapper') : NULL;
    // Merge with static data, e.g. provided by preprocess hooks.
    if ($static_token_data = &drupal_static('mustache_tokens', [])) {
      foreach ($static_token_data as $type => $value) {
        if (isset($value) && !isset($token_data[$type])) {
          $token_data[$type] = $value;
        }
      }
    }

    // Merge the token data with any global context that provides a value.
    $context_repository = $this->contextRepository;
    if (!isset(static::$contextIds)) {
      static::$contextIds = array_keys($context_repository->getAvailableContexts());
    }
    if (!empty(static::$contextIds)) {
      foreach ($context_repository->getRuntimeContexts(static::$contextIds) as $context) {
        $context_data_type = explode(':', $context->getContextDefinition()->getDataType());
        // For "entity:*" data types, we skip its prefixes and only respect
        // the last entry of the data type string (e.g. for "entity:node", we
        // want to only use "node").
        $type = end($context_data_type);
        $value = $context->getContextValue();
        if (isset($value) && !isset($token_data[$type])) {
          $token_data[$type] = $value;
        }
      }
    }

    // Filter out any data that is not allowed to be viewed.
    // Also re-map entities that may have a different token type defined.
    foreach ($token_data as $type => $value) {
      if ($value instanceof CacheableDependencyInterface) {
        $bubbleable_metadata->addCacheableDependency($value);
      }
      if ($check_access && $value instanceof AccessibleInterface) {
        /** @var \Drupal\Core\Access\AccessResultInterface $access_result */
        $access_result = $value->access($check_access, NULL, TRUE);
        $bubbleable_metadata->addCacheableDependency($access_result);
        if (!$access_result->isAllowed()) {
          unset($token_data[$type]);
          continue;
        }
      }
      if (!($value instanceof EntityInterface)) {
        continue;
      }
      /** @var \Drupal\Core\Entity\EntityInterface $value */
      if ($entity_token_type_mapper) {
        $the_real_token_type = $entity_token_type_mapper->getTokenTypeForEntityType($value->getEntityTypeId(), TRUE);
      }
      else {
        $definition = $this->entityTypeManager->getDefinition($value->getEntityTypeId());
        $the_real_token_type = $definition->get('token_type') ?: $value->getEntityTypeId();
      }
      if (!isset($token_data[$the_real_token_type])) {
        $token_data[$the_real_token_type] = $value;
      }
    }
  }

  /**
   * Get the token iterate service.
   *
   * @return \Drupal\mustache\MustacheTokenIterate
   *   The token iterate service.
   */
  protected static function getTokenIterate() {
    return \Drupal::service('mustache.token_iterate');
  }

}
