<?php

namespace Drupal\mustache\Render;

use Drupal\Component\Render\MarkupInterface;

/**
 * Passes multiple safe strings through the render system.
 *
 * This iterable markup allows for both concatenating the contained items,
 * but also define a value that may be used instead of concatenating the items.
 * For nesting iterable markup items, a child markup item may reference to its
 * parent item through its "parent" property.
 *
 * This object should only be constructed with known safe strings. If there is
 * any risk that the string contains user-entered data that has not been
 * filtered first, it must not be used.
 *
 * @internal
 *   This class must only be used inside the mustache_templates project.
 */
final class IterableMarkup implements MarkupInterface, \Countable, \IteratorAggregate, \ArrayAccess {

  /**
   * The iterable items.
   *
   * @var array
   */
  public $items;

  /**
   * The string value to use instead of concatenating the items.
   *
   * @var mixed
   */
  public $value;

  /**
   * The parent.
   *
   * @var mixed
   */
  public $parent;

  /**
   * The separator to use when concatenating as string.
   *
   * @var string
   */
  public $separator;

  /**
   * The position of this item within its parent's list of items.
   *
   * Position counting starts at value 1.
   *
   * @var int
   */
  public $position = 1;

  /**
   * Whether this item is the first one within its parent's list of items.
   *
   * @var bool
   */
  public $first = TRUE;

  /**
   * Whether this item is the last one within its parent's list of items.
   *
   * @var bool
   */
  public $last = TRUE;

  /**
   * Create an iterable markup object.
   *
   * @param array $items
   *   The items to iterate on.
   * @param mixed $value
   *   (Optional) The string value to use instead of concatenating the items.
   * @param mixed $parent
   *   (Optional) The parent, usually another IterableMarkup instance.
   * @param string $separator
   *   (Optional) The separator to use on string concatenation.
   *
   * @return static
   *   The iterable markup object.
   */
  public static function create(array $items = [], $value = NULL, $parent = NULL, $separator = ', ') {
    $instance = new static();
    $instance->items = $items;
    $instance->value = $value;
    $instance->parent = $parent;
    $instance->separator = $separator;
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function __toString() {
    if ($this->value !== NULL) {
      return (string) $this->value;
    }
    return implode($this->separator, $this->items);
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return [
      'value' => $this->__toString(),
      'items' => $this->items,
      'separator' => $this->separator,
      'position' => $this->position,
      'first' => $this->first,
      'last' => $this->last,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function count() {
    if ($this->items) {
      return count($this->items);
    }
    if ($this->value !== NULL) {
      return 1;
    }
    return 0;
  }

  /**
   * {@inheritdoc}
   */
  public function getIterator() {
    if ($this->items || $this->value === NULL) {
      return new \ArrayIterator($this->items);
    }
    return new \ArrayIterator([&$this->value]);
  }

  /**
   * {@inheritdoc}
   */
  public function offsetSet($offset, $value) {
    if (is_null($offset)) {
      $this->items[] = $value;
    }
    else {
      $this->items[$offset] = $value;
    }
    $this->updatePositions();
  }

  /**
   * {@inheritdoc}
   */
  public function offsetExists($offset) {
    return isset($this->items[$offset]);
  }

  /**
   * {@inheritdoc}
   */
  public function offsetUnset($offset) {
    unset($this->items[$offset]);
    $this->updatePositions();
  }

  /**
   * {@inheritdoc}
   */
  public function &offsetGet($offset) {
    if (isset($this->items[$offset])) {
      return $this->items[$offset];
    }
    return NULL;
  }

  /**
   * Implements the magic __get() method.
   *
   * @return mixed
   *   The value.
   */
  public function &__get($name) {
    if (isset($this->items[$name])) {
      return $this->items[$name];
    }
    foreach ($this->items as $item) {
      if (isset($item[$name])) {
        return $item[$name];
      }
    }
    return NULL;
  }

  /**
   * Implements the magic __get() method.
   *
   * @return bool
   *   The value.
   */
  public function __isset($name) {
    if (isset($this->items[$name])) {
      return TRUE;
    }
    if (!empty($this->items)) {
      foreach ($this->items as $item) {
        if (isset($item[$name])) {
          return TRUE;
        }
      }
    }
    return FALSE;
  }

  /**
   * Updates the positions on the child items.
   */
  public function updatePositions() {
    reset($this->items);
    $first = TRUE;
    $i = 0;
    foreach ($this->items as &$item) {
      $i++;
      if ($item instanceof static) {
        $item->position = $i;
        $item->first = $first;
        $item->last = FALSE;
      }
      elseif (is_array($item)) {
        $item['position'] = $i;
        $item['first'] = $first;
        $item['last'] = FALSE;
      }
      $first = FALSE;
    }
    if ($item instanceof static) {
      $item->last = TRUE;
    }
    elseif (is_array($item)) {
      $item['last'] = TRUE;
    }
  }

}
